@@ -22,14 +22,14 @@ The "supervisor" command suports several modes of operation:
 
 """
 
-from __future__ import absolute_import
-from __future__ import with_statement
+from __future__ import absolute_import, with_statement
 
 import sys
 import os
 import time
 from optparse import make_option
 from textwrap import dedent
+import traceback
 from ConfigParser import RawConfigParser, NoOptionError
 try:
     from cStringIO import StringIO
@@ -38,9 +38,13 @@ except ImportError:
 
 from supervisor import supervisord, supervisorctl
 
+from watchdog.observers import Observer
+from watchdog.observers.polling import PollingObserver
+
 from django.core.management.base import BaseCommand, CommandError
 
 from djsupervisor.config import get_merged_config
+from djsupervisor.events import CallbackModifiedHandler
 
 
 class Command(BaseCommand):
@@ -184,7 +188,7 @@ class Command(BaseCommand):
         """Command 'supervisor getconfig' prints merged config to stdout."""
         if args:
             raise CommandError("supervisor getconfig takes no arguments")
-        print cfg_file.getvalue()
+        print cfg_file.read()
         return 0
 
     def _handle_autoreload(self,cfg_file,*args,**options):
@@ -201,16 +205,56 @@ class Command(BaseCommand):
         if args:
             raise CommandError("supervisor autoreload takes no arguments")
         live_dirs = self._find_live_code_dirs()
-        mtimes = {}
         reload_progs = self._get_autoreload_programs(cfg_file)
-        while True:
-            if self._code_has_changed(live_dirs,mtimes):
-                #  Fork a subprocess to make the restart call.
-                #  Otherwise supervisord might kill us and cancel the restart!
-                if os.fork() == 0:
-                    self.handle("restart",*reload_progs,**options)
-                return 0
-            time.sleep(1)
+
+        observer = Observer()
+
+        def autoreloader():
+            """
+            Forks a subprocess to make the restart call.
+            Otherwise supervisord might kill us and cancel the restart!
+            """
+            if os.fork() == 0:
+                self.handle("restart", *reload_progs, **options)
+
+        handler = CallbackModifiedHandler(callback=autoreloader,
+                                          patterns=['*.py', '*.pyc', '*.pyo'],
+                                          ignore_patterns=[".*", "#*", "*~"],
+                                          ignore_directories=True)
+
+        # Try to add watches using the platform-specific observer.
+        # If this fails, print a warning and fall back to the PollingObserver.
+        # This will avoid errors with e.g. too many inotify watches.
+        observer = None
+        for ObserverCls in (Observer, PollingObserver):
+            observer = ObserverCls()
+            try:
+                for live_dir in set(live_dirs):
+                    observer.schedule(handler, live_dir, True)
+                break
+            except Exception:
+                print>>sys.stderr, "COULD NOT WATCH FILESYSTEM USING"
+                print>>sys.stderr, "OBSERVER CLASS: ", ObserverCls
+                traceback.print_exc()
+                observer.start()
+                observer.stop()
+
+        # Fail out if none of the observers worked.
+        if observer is None:
+            print>>sys.stderr, "COULD NOT WATCH FILESYSTEM"
+            return 1
+
+        # Poll of we have an observer.
+        # TODO: Is this sleep necessary?  Or will it suffice
+        # to block indefinitely on something and wait to be killed?
+        observer.start()
+        try:
+            while True:
+                time.sleep(1)
+        except KeyboardInterrupt:
+            observer.stop()
+        observer.join()
+        return 0
 
     def _get_autoreload_programs(self,cfg_file):
         """Get the set of programs to auto-reload when code changes.
@@ -231,29 +275,12 @@ class Command(BaseCommand):
                     pass
         return reload_progs
 
-    def _code_has_changed(self,live_dirs,mtimes):
-        """Check whether code under the given directories has changed.
-
-        This is a simple check based on file mtime.  New or deleted files
-        don't count as code changes.
-        """
-        for filepath in self._find_live_code_files(live_dirs):
-            try:
-                stat = os.stat(filepath)
-            except EnvironmentError:
-                continue
-            if filepath not in mtimes:
-                mtimes[filepath] = stat.st_mtime
-            else:
-                if mtimes[filepath] != stat.st_mtime:
-                    return True
-
     def _find_live_code_dirs(self):
         """Find all directories in which we might have live python code.
 
         This walks all of the currently-imported modules and adds their
         containing directory to the list of live dirs.  After normalization
-        and de-duplication, we get a pretty good approximation of the 
+        and de-duplication, we get a pretty good approximation of the
         directories on sys.path that are actively in use.
         """
         live_dirs = []
@@ -282,23 +309,6 @@ class Command(BaseCommand):
                 live_dirs.append(dirnm)
         return live_dirs
 
-    def _find_live_code_files(self,live_dirs):
-        """Find live python code files, that must be watched for changes.
-
-        Given a pre-computed list of directories in which to search, this
-        method finds all the python files under those directories and yields
-        their full paths.
-        """
-        for dirnm in live_dirs:
-            for (subdirnm,_,filenms) in os.walk(dirnm):
-                for filenm in filenms:
-                    try:
-                        filebase,ext = filenm.rsplit(".",1)
-                    except ValueError:
-                        continue
-                    if ext in ("py","pyc","pyo",):
-                        yield os.path.join(subdirnm,filebase+".py")
-
 
 class OnDemandStringIO(object):
     """StringIO standin that demand-loads its contents and resets on EOF.
